
def i2s(v, chars):
    N = len(chars)
    base = 1
    bases = []
    ceiling = 1
    while ceiling <= v:
        bases.append(base)
        base *= N
        ceiling += base

    chrs = ''
    n = v - ceiling
    for base in reversed(bases):
        i = n // base
        n -= i*base
        ch = chars[i]
        chrs += ch

    return chrs


def s2i(s, ints):
    sz = len(s)
    N = len(ints)
    n = (N**sz-1)//(N-1)
    base = 1
    for ch in reversed(s):
        i = ints[ch]
        n += i*base
        base *= N
    return n


def b2s(b, chars):
    n = 0
    for byte in b:
        n *= 256
        n += byte
    return i2s(n, chars)


class Converter:
    '''Base-class for converters between non-negative integers and strings comprised of a given alphabet

    See "base" 62, 64, and 90 versions below, but note that they work differently than e.g. standard base64; these
    encodings are isomorphisms between all natural numbers and all strings:
    - 0 ⟺ ''
    - [1,N) ⟺ <1-char strings>
    - [N, N²+N) ⟺ <2-char strings>
    - [N²+N, N³+N²+N) ⟺ <3-char strings>
    - etc.
    '''
    I2S = None

    def __init__(self):
        self.S2I = {ch:i for i, ch in enumerate(self.I2S)}

    def __call__(self, v):
        if isinstance(v, str):
            return s2i(v, self.S2I)
        elif isinstance(v, int):
            return i2s(v, self.I2S)
        elif isinstance(v, bytes):
            return b2s(v, self.I2S)
        else:
            raise ValueError(f'Unrecognized type ({type(v)}): {v}')


class B62(Converter): I2S = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
b62 = B62()


class B64(Converter): I2S = '+/0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
b64 = B64()


class B90(Converter): I2S = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz'
b90 = B90()




from .. import process
from ..process import output


def msg(ref=None):
    return output('git','log','-n1','--format=%B',ref).decode().strip()


def fmt(*refs: str, fmt: str = '%h', **kwargs) -> str:
    return process.line('git', 'log', '-1', f'--format={fmt}', *refs, **kwargs)


def sha(ref: str, **kwargs) -> str:
    return fmt(ref, fmt='%H', **kwargs)


def short_sha(ref: str, **kwargs) -> str:
    return fmt(ref, fmt='%h', **kwargs)



from functools import wraps
from os import environ
import re
from sys import stderr
from typing import Optional, Literal
from utz import process

GITHUB_HTTPS_URL_RGX = r'https://github.com/(?P<nameWithOwner>[^/]+/[^/]+?)(?:\.git)?'
GITHUB_SSH_URL_RGX = r'git@github.com:(?P<nameWithOwner>[^/]+/[^/]+?)(?:\.git)?'


def parse_url(url: str, err: Literal['raise', 'stderr', 'none'] = 'raise') -> Optional[str]:
    m = re.fullmatch(GITHUB_HTTPS_URL_RGX, url)
    if m:
        return m.group('nameWithOwner')
    m = re.fullmatch(GITHUB_SSH_URL_RGX, url)
    if m:
        return m.group('nameWithOwner')
    if err == 'raise':
        raise ValueError(f'Could not parse GitHub url: {url}')
    elif err == 'stderr':
        print(f'Could not parse GitHub url: {url}', file=stderr)
    return None


GITHUB_REPOSITORY = 'GITHUB_REPOSITORY'


def repository_option(
        *flag_args,
        env=GITHUB_REPOSITORY,
        help='Repository name (with owner, e.g. "owner/repo"), defaults to $GITHUB_REPOSITORY then `gh repo view --json nameWithOwner`', **flag_kwargs,
):
    if not flag_args:
        flag_args = ('-R', '--repository')

    def option(fn):
        import click
        @click.option(*flag_args, help=help, **flag_kwargs)
        @wraps(fn)
        def _fn(*args, repository=None, **kwargs):
            if not repository:
                repository = environ.get(env)
                if not repository:
                    verbose = kwargs.get('verbose', 0)
                    repository = process.json('gh', 'repo', 'view', '--json', 'nameWithOwner', log=None if verbose else False)['nameWithOwner']
            return fn(*args, repository=repository, **kwargs)

        return _fn

    return option



from os import getcwd
from os.path import basename, exists
import re
import setuptools
from setuptools import find_packages

from utz.process import line
from utz.version import git_version


class Compute:
    """Container for computing fallback values to pass to setup()"""

    def long_description(self):
        """Read in README.md as the `long_description`"""
        if exists('README.md'):
            with open('README.md', "r") as fh:
                return fh.read()

    def long_description_content_type(self):
        """Default to markdown format for `long_description`"""
        return "text/markdown"

    def install_requires(self):
        """Read in requirements.txt as `install_requires`"""
        if exists('requirements.txt'):
            with open('requirements.txt', "r") as fh:
                return fh.read()

    def version(self):
        return git_version()

    def name(self):
        """`name` defaults to the name of the containing directory"""
        return basename(getcwd())

    def author(self):
        """Infer author name from most recent commit"""
        return line('git', 'log', '-n1', '--format=%an')

    def author_email(self):
        """Infer author email from most recent commit"""
        return line('git', 'log', '-n1', '--format=%ae')

    def description(self):
        """Set `description` to the contents of a <p> first-child of an initial <h1>"""
        try:
            import lxml
            import mistune
            html = mistune.html(self.long_description())

            from lxml.html import fragments_fromstring
            [ h1, p, *_ ] = fragments_fromstring(html)
            if h1.tag != 'h1' or p.tag != 'p':
                raise ValueError('Expected initial <h1> followed by <p> while parsing `description` from README.md')

            return p.text_content()
        except ImportError:
            md = self.long_description()
            md_lines = md.split('\n')
            for line in md_lines:
                if not re.match(r'[#\[!\-|<]', line):
                    return line
            return None

    def packages(self):
        return find_packages()

    def classifiers(self):
        classifiers = [
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ]

        license = self.license()
        if license:
            classifiers += [f"License :: OSI Approved :: {license} License"]

        return classifiers

    def python_requires(self): return '>=3.8'

    def license(self):
        if exists('LICENSE'):
            with open('LICENSE', 'r') as f:
                lines = f.readlines()
                first_lines = [ line.strip() for line in lines[:2] ]
                if first_lines == ['Apache License', 'Version 2.0, January 2004',]: return 'Apache v2'
                if first_lines == ['MIT License', '',]: return 'MIT'


def setup(**kwargs):
    """Wrapper for setuptools.setup(), with convenient fallback values for common keys."""
    c = Compute()

    def compute(*keys):
        for k in keys:
            if k not in kwargs:
                fn = getattr(c, k)
                v = fn()
                s = str(v)
                s = s[:100] + '…' if len(s) > 100 else s
                print('Fallback value: %s = %s' % (k, s))
                kwargs[k] = v

    compute(
        'name',
        'version',
        'description',
        'long_description',
        'long_description_content_type',
        'install_requires',
        'license',
        'author',
        'author_email',
        'packages',
        'classifiers',
        'python_requires',
    )

    setuptools.setup(**kwargs)



from functools import partial
from io import TextIOWrapper
from os import environ as env
from os.path import exists, join
from sys import stderr
from typing import Union, Literal

from IPython.display import Image

import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.validators.scatter.marker import SymbolValidator
symbols = SymbolValidator().values[2::12]

from utz import err


# Env vars to fall back to, for various plot kwargs
PLOT_SHOW_VAR = 'UTZ_PLOT_SHOW'
PLOT_DIR_VAR = 'UTZ_PLOT_DIR'
PLOT_MARGIN_VAR = 'UTZ_PLOT_MARGIN'
PLOT_BG_VAR = 'UTZ_PLOT_BG'
PLOT_GRID_VAR = 'UTZ_PLOT_GRID'
PLOT_SUBTITLE_SIZE_VAR = 'UTZ_PLOT_SUBTITLE_SIZE'

DEFAULT_BG = 'white'
DEFAULT_SHOW = 'html'
DEFAULT_GRID = '#ccc'
DEFAULT_SUBTITLE_SIZE = '0.8em'


class Unset:
    pass


_Unset = type(Unset)


Title = str | list[str] | None


def title(s: Title, subtitle_size: str | None = None) -> str | None:
    if isinstance(s, list):
        if subtitle_size is None:
            subtitle_size = env.get(PLOT_SUBTITLE_SIZE_VAR, DEFAULT_SUBTITLE_SIZE)

        title, *subtitles = s
        prefix = f'<br><span style="font-size:{subtitle_size}">'
        suffix = '</span>'
        title = prefix.join([title] + [ f'{subtitle}{suffix}' for subtitle in subtitles ])
        return title
    else:
        return s


make_title = title


def save(fig, title, name, **kwargs):
    return plot(fig, title, name=name, **kwargs)


def plot(
        fig: Figure,
        title: Title,
        name: str | None = None,
        bg: str | None | _Unset = Unset,
        subtitle_size: str | None = None,
        hoverx: bool = False,
        hovertemplate: Title = None,
        png_title: bool = True,
        yrange: str | None = 'tozero',
        legend: Union[bool, dict, None] = None,
        bottom_legend: Union[bool, Literal['all']] = False,
        pretty: bool = False,
        margin: Union[None, int, dict, _Unset] = Unset,
        dir: str | None = None,
        w: int | None = None,
        h: int | None = None,
        xtitle: Title = None,
        ytitle: Title = None,
        ltitle: Title = None,
        grid: str | None | _Unset = Unset,
        xgrid: str | None = None,
        ygrid: str | None = None,
        x: dict | str | None = None,
        y: dict | str | None = None,
        log: TextIOWrapper | bool | None = stderr,
        show: Literal['png', 'file', 'html'] | bool | None = None,
        zerolines: bool | Literal['x', 'y'] | None = True,
        **layout,
):
    """Plotly wrapper with convenience kwargs for common configurations.

    Args:
        fig: Plotly figure
        title: Title string or list of strings (tail will be converted to "subtitle" lines)
        name: "Stem" for output file paths
        bg: Background color, falls back to $UTZ_PLOT_BG, defaults to `DEFAULT_BG` ("white")
        subtitle_size: Subtitle font size; falls back to $UTZ_PLOT_SUBTITLE_SIZE, defaults to "0.8em"
        hoverx: Alias for `hovermode="x"`
        hovertemplate: Hover template; `list[str]` will be `<br>.join()`'d.
        png_title: Whether to include the title in PNG (sometimes nice to disable this, e.g. for PNGs that will be embedded in Markdown that includes the title).
        yrange: Y-axis range mode
        legend: Legend configuration
        bottom_legend: Orient the legend horizontally along the bottom; True ⇒ JSON/PNG outputs only, 'all' ⇒ also in returned Figure
        pretty: Pretty-print JSON output
        margin: Margin size or dict of sizes (falls back to $UTZ_PLOT_MARGIN)
        dir: Output directory (falls back to $UTZ_PLOT_DIR)
        w: Width; applied to Figure and any resulting PNG output
        h: Height; applied to Figure and any resulting PNG output
        xtitle: X-axis title; string or list of strings (tail will be converted to "subtitle" lines)
        ytitle: Y-axis title; string or list of strings (tail will be converted to "subtitle" lines)
        ltitle: Legend title; string or list of strings (tail will be converted to "subtitle" lines)
        grid: Grid color (x- and y-axes)
        xgrid: X-axis grid color
        ygrid: Y-axis grid color
        x: X-axis configuration (passed to `update_xaxes`)
        y: Y-axis configuration (passed to `update_yaxes`)
        log: Whether/where to print log info (e.g. about files written); defaults to stderr
        show: Format to return the Figure in: 'png' ⇒ return PNG `Image`, 'file' ⇒ return `Image(filename=…)` pointing to output `.png`, 'html' ⇒ return default Plotly Figure HTML, False ⇒ None (don't show figure, if `plot(…)` call is last expression in a notebook cell). Falls back to $UTZ_PLOT_SHOW, defaults to 'html'.
        zerolines: Whether to show zero lines; 'x' ⇒ only on x-axis, 'y' ⇒ only on y-axis, True (default) ⇒ on both axes
    """
    mk_title = partial(make_title, subtitle_size=subtitle_size)
    xtitle = mk_title(xtitle)
    if xtitle:
        fig.update_xaxes(title=dict(text=xtitle))

    ytitle = mk_title(ytitle)
    if ytitle:
        fig.update_yaxes(title=dict(text=ytitle))

    ltitle = mk_title(ltitle)
    layout['legend_title'] = layout.get('legend_title', ltitle or '')

    if w is not None:
        layout['width'] = w
    if h is not None:
        layout['height'] = h

    if bg is Unset:
        bg = env.get(PLOT_BG_VAR)
        if bg is None:
            bg = DEFAULT_BG

    if bg:
        layout['plot_bgcolor'] = bg
        layout['paper_bgcolor'] = bg

    if isinstance(hovertemplate, list):
        hovertemplate = '<br>'.join(hovertemplate)
    if hoverx:
        layout['hovermode'] = 'x'
        fig.update_traces(hovertemplate=hovertemplate)
    elif hovertemplate:
        fig.update_traces(hovertemplate=hovertemplate)

    if yrange:
        layout['yaxis_rangemode'] = yrange

    bottom_legend_kwargs = dict(
        orientation='h',
        x=0.5,
        xanchor='center',
        yanchor='top',
    ) if bottom_legend else {}
    if bottom_legend == 'all':
        if 'legend' not in layout:
            layout['legend'] = {}
        layout['legend'].update(**bottom_legend_kwargs)

    if legend is False:
        layout['showlegend'] = False
    elif isinstance(legend, dict):
        layout['legend'] = legend

    if grid is Unset:
        grid = env.get(PLOT_GRID_VAR)
        if grid is None:
            grid = DEFAULT_GRID

    if xgrid:
        fig.update_xaxes(gridcolor=xgrid)
    elif grid:
        fig.update_xaxes(gridcolor=grid)

    if ygrid:
        fig.update_yaxes(gridcolor=ygrid)
    elif grid:
        fig.update_yaxes(gridcolor=grid)

    if zerolines == 'x' or zerolines is True:
        fig.update_xaxes(
            zeroline=True,
            zerolinecolor=xgrid or grid,
            zerolinewidth=1,
        )

    if zerolines == 'y' or zerolines is True:
        fig.update_yaxes(
            zeroline=True,
            zerolinecolor=ygrid or grid,
            zerolinewidth=1,
        )

    if isinstance(x, str):
        fig.update_xaxes(title_text=x)
    elif isinstance(x, dict):
        fig.update_xaxes(**x)

    if isinstance(y, str):
        fig.update_yaxes(title_text=y)
    elif isinstance(y, dict):
        fig.update_yaxes(**y)

    title = mk_title(title)
    title_layout = dict(title_text=title, title_x=0.5) if title else {}
    if png_title:
        layout.update(title_layout)
    fig.update_layout(**layout)

    saved = go.Figure(fig)
    if not png_title:
        # only need to do this if it wasn't already done above
        fig.update_layout(**title_layout)

    if bottom_legend is True:
        saved.update_layout(
            legend=bottom_legend_kwargs,
        )

    if margin is Unset:
        margin = env.get(PLOT_MARGIN_VAR)
        if margin:
            margin = json.loads(margin)
    if isinstance(margin, int):
        margin = { k: margin for k in 'trbl' }
    if margin:
        saved.update_layout(margin=margin)

    if log is True:
        log = err
    elif log:
        log = partial(print, file=log)
    else:
        log = lambda: ()

    if name:
        if dir is None:
            dir = env.get(PLOT_DIR_VAR)
        if dir:
            if not exists(dir):
                makedirs(dir, exist_ok=True)
        else:
            dir = '.'
        json_path = join(dir, f'{name}.json')
        saved.write_json(json_path, pretty=pretty)
        png_path = join(dir, f'{name}.png')
        log(f"Wrote plot JSON to {png_path}")
        saved.write_image(png_path)
        log(f"Wrote plot image to {png_path}")

    if show is None:
        show = env.get(PLOT_SHOW_VAR)
        if show is None:
            show = DEFAULT_SHOW
    if show == 'file':
        return Image(filename=png_path)
    elif show == 'png':
        return Image(fig.to_image())
    elif show is None or show is False:
        return
    else:
        return fig



from subprocess import CalledProcessError

from ..process import *
from . import diff


def mv(old, new, untracked=False, checkout=False):
    if current() == old:
        if diff.exists(untracked=untracked):
            raise ValueError(f"Local changes found; refusing to reset --hard")
        else:
            run('git','reset','--hard',new)
    else:
        run('git','branch','-f',old,new)

    if checkout:
        run('git','checkout',new)

def ls():
    return lines('git','for-each-ref','--format=%(refname:short)','refs/heads')
        
def exists(name):
    return check('git','show-ref','--verify','--quiet',f'refs/heads/{name}')

def current(sha_ok=False):
    if sha_ok:
        try:
            return line('git','symbolic-ref','-q','--short','HEAD')
        except CalledProcessError:
            return line('git','log','--no-walk','--format=%h')
    elif sha_ok is None:
        return line('git','symbolic-ref','-q','--short','HEAD', err_ok=True)
    else:
        return line('git','symbolic-ref','-q','--short','HEAD')




from functools import cached_property
from inspect import isfunction


class Methods:
    def __init__(self, o):
        klz = type(o)
        klzm = klz.__dict__
        self.methods = set()
        self.cmethods = set()
        self.smethods = set()
        self.smembers = set()
        self.properties = set()
        self.cached_properties = set()

        for k,v in klzm.items():
            if k.startswith('_'): continue
            if isfunction(v):
                self.methods.add(k)
            elif isinstance(v, classmethod):
                self.cmethods.add(k)
            elif isinstance(v, staticmethod):
                self.smethods.add(k)
            elif isinstance(v, property):
                self.properties.add(k)
            elif isinstance(v, cached_property):
                self.cached_properties.add(k)
            else:
                self.smembers.add(k)

        self.ivars = set(k for k in dir(o) if not k.startswith('_')).difference(klzm.keys())



from functools import partial
# # Pandas imports / aliases / helpers

from pathlib import Path

from os import remove
from os.path import exists, isdir, splitext

from utz.imports import _try
with _try:
    from dateutil.parser import parse
with _try:
    from numpy import nan, array, ndarray
    import pandas as pd
    from pandas import \
        concat, \
        DataFrame as DF, \
        Series, \
        isna, \
        read_csv, read_excel, read_json, read_parquet, read_sql, read_sql_query, read_sql_table, \
        date_range, to_datetime as to_dt, Timedelta as Δ, NaT, \
        get_option, set_option

    def display(r=None, c=None):
        """Set the default number of rows and columns for Pandas to display"""
        if r:
            pd.options.display.max_rows = r
        if c:
            pd.options.display.max_columns = c

    display(100, 100)


from shutil import rmtree

# ## crosstab
xtab = pd.crosstab
xtabn = partial(pd.crosstab, dropna=False)

# ## value_counts
def vc(p, **kwargs):
    return p.value_counts(**kwargs)


def vcs(p, **kwargs):
    return p.value_counts(**kwargs).sort_index()


def vcn(p, **kwargs):
    return p.value_counts(dropna=False, **kwargs)


def vcns(p, **kwargs):
    return p.value_counts(dropna=False, **kwargs).sort_index()


def vcvc(p, **kwargs):
    return p.value_counts(**kwargs).value_counts()


def vcvcs(p, **kwargs):
    return p.value_counts(**kwargs).value_counts().sort_index()


# ## Concat / Creation

def sxs(*dfs, **kwargs):
    """Concat some DataFrames "side by side\""""
    return concat(dfs, axis=1, **kwargs)


# ## Counting

def series_counts(series, ascending = None):
    df_new = DF(series.value_counts())
    column_name = df_new.columns[0]
    df_new['index'] = df_new.index
    ascending = [False,True] if ascending is None else ascending
    return df_new.sort_values(by = [column_name,'index'], ascending = ascending)[column_name]


def sort_by_series(df, series):
    tmp_col_name = series.name or '_'
    while tmp_col_name in df:
        tmp_col_name=f'_{tmp_col_name}'

    if isinstance(series, str):
        series = df[series]

    df2 = df.copy()
    df2[tmp_col_name] = series
    return         df2         .sort_values(tmp_col_name)         .drop(columns=tmp_col_name)


# ## File paths

def file_dict(path: Path):
    name = path.name
    pieces = name.rsplit('.', 1)
    if len(pieces) == 2:
        base, xtn = pieces
    else:
        [ base ] = pieces
        xtn = nan
    
    return dict(parent=path.parent, name=name, base=base, xtn=xtn, exists=path.exists())


def annotate_files(files):
    if files.empty:
        dtypes = {
            'parent': object,
            'name': str,
            'base': str,
            'xtn': str,
            'exists': bool,
        }
        return DF([], index=files, columns=dtypes.keys()).astype(dtypes)
    else:
        extra = files.apply(file_dict).apply(Series)
        files = sxs(files, extra).set_index(files.name)
        return files


def load_files(files, glob=None, name=None):
    if isinstance(files, Path):
        name = name or 'path'
        files = Series(files.glob(glob), name=name, dtype=object)
    else:
        name = name or files.name
    files = annotate_files(files)
    return files


# ## Dates

def expand_date(s):
    d = parse(s).date()
    return dict(date=d, year=d.year, month=d.month, day=d.day)


def to_parquet(df, out_path, verify_extension=True, *args, **kwargs):
    if verify_extension:
        _, extension = splitext(out_path)
        if extension != '.parquet':
            raise Exception(f"Refusing to write parquet dataset to non-parquet path {out_path}")

    if exists(out_path):
        if isdir(out_path):
            rmtree(out_path)
        else:
            remove(out_path)

    return df.to_parquet(out_path, *args, **kwargs)
